import logging
from flask import g

from flask_logger_decorator.utils import generate_request_id
from flask_logger_decorator.parser import auto_parser
from flask_logger_decorator.config import config
from flask_logger_decorator.request_decorators import get_request_trace_info, get_response_trace_info
from flask_logger_decorator.logger import debug


class LoggerExtension(object):

    def __init__(self, app=None):
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        if not hasattr(app, 'extensions'):   # pragma: no cover
            app.extensions = {}
        app.extensions['flask-logger-decorator-extended'] = self

        self._set_default_configuration_options(app)

    def _set_default_configuration_options(self, app):
        app.config.setdefault('LOG_REQUEST_ID_G_OBJECT_ATTRIBUTE', 'flask_log_request_id')
        app.config.setdefault('LOG_LEVEL', logging.DEBUG)
        app.config.setdefault('LOG_ALL_REQUESTS', True)

        @app.before_request
        def _persist_request_id():
            g_object_attr = config.log_request_id

            setattr(g, g_object_attr, auto_parser())
            if g.get(g_object_attr) is None:
                setattr(g, g_object_attr, generate_request_id())

            debug(config.request_log_name, get_request_trace_info())

        app.after_request(self._log_http_event)

    @staticmethod
    def _log_http_event(response):
        debug(config.request_log_name, get_response_trace_info(response))
        return response



from flask import request

_parsers = []


def request_id_parser(fn):
    _parsers.append(fn)
    return fn


def generic_http_header_parser_for(header_name):
    """
    A parser factory to extract the request id from an HTTP header
    :return: A parser that can be used to extract the request id from the current request context
    :rtype: ()->str|None
    """

    def parser():
        request_id = request.headers.get(header_name, '').strip()

        if not request_id:
            # If the request id is empty return None
            return None
        return request_id
    return parser


@request_id_parser
def x_request_id():
    """
    Parser for generic X-Request-ID header
    :rtype: str|None
    """
    return generic_http_header_parser_for('X-Request-ID')()


@request_id_parser
def x_correlation_id():
    """
    Parser for generic X-Correlation-ID header
    :rtype: str|None
    """
    return generic_http_header_parser_for('X-Correlation-ID')()


@request_id_parser
def amazon_elb_trace_id():
    """
    Get the amazon ELB trace id from current Flask request context
    :return: The found Trace-ID or None if not found
    :rtype: str | None
    """
    amazon_request_id = request.headers.get('X-Amzn-Trace-Id', '')
    trace_id_params = dict(x.split('=') if '=' in x else (x, None) for x in amazon_request_id.split(';'))
    if 'Self' in trace_id_params:
        return trace_id_params['Self']
    if 'Root' in trace_id_params:
        return trace_id_params['Root']

    return None


def auto_parser():
    """
    Meta parser that will try all known parser and it will bring the first found id
    :param list[Callable] parsers: A list of callable parsers to try to extract request_id
    :return: The request id if it is found or None
    :rtype: str|None
    """

    for parser in _parsers:
        request_id = parser()
        if request_id is not None:
            return request_id

    return None  # request-id was not found



import logging
import sys
from flask_logger_decorator.config import config

FLASK_LOGGER = {}


def _setup_stdout(logger):
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(config.log_level)

    formatter = logging.Formatter(config.log_format_default, config.log_format_time_default)
    stdout_handler.setFormatter(formatter)
    logger.addHandler(stdout_handler)


def get_logger(name=None):
    logger = FLASK_LOGGER.get(name)
    if not logger:
        logger = logging.getLogger(name)
        logger.setLevel(config.log_level)
        logger.handlers = []
        _setup_stdout(logger)
    return logger


def _log(name):
    logger = FLASK_LOGGER.get(name)
    if not logger:
        logger = get_logger(name)
        FLASK_LOGGER[name] = logger
    return logger


def debug(name, message, extra=None):
    _log(name).debug(message, extra=extra)


def info(name, message, extra=None):
    _log(name).info(message, extra=extra)


def warning(name, message, extra=None):
    _log(name).warning(message, extra=extra)


def error(name, message, extra=None):
    _log(name).error(message, extra=extra)


def critical(name, message, extra=None):
    _log(name).critical(message, extra=extra)



from functools import wraps
from reprlib import Repr
import inspect
from flask import request, g
from flask_logger_decorator.config import config
from flask_logger_decorator.logger import debug

__r = Repr()
__r.maxarray = __r.maxarray * 10
__r.maxdict = __r.maxdict * 10
__r.maxstring = __r.maxstring * 10


def request_tracing(fn):
    """
    A decorator to tracing request.
    """
    @wraps(fn)
    def wrapper(*args, **kwargs):
        tracing_request(fn, *args, **kwargs)
        return fn(*args, **kwargs)
    return wrapper


def tracing_request(fn, *args, **kwargs):
    function_args = ' '.join(_get_fn_args(fn, *args, **kwargs))
    trace_info = ' '.join(_get_fn_extra_info(fn))
    func_msg = 'func_name:{} func_args:{} trace_info:{}'.format(
        fn.__name__, function_args, trace_info)
    request_msg = get_request_trace_info()
    debug(config.request_log_name, 'req_msg={}; func_msg={}'.format(request_msg, func_msg))


def _get_fn_args(fn, *args, **kwargs):
    traced_func_args = []
    for k, v in inspect.getcallargs(fn, *args, **kwargs).items():
        if k == 'self':
            continue

        v = '********' if k in {'password', 'secret'} else __r.repr(v).replace('"', "'")
        traced_func_args.append('{}={}'.format(k, v))
    return traced_func_args


def _get_fn_extra_info(fn):
    trace_info_list = ['trace_pathname={}'.format(inspect.getfile(fn))]
    try:
        trace_info_list.append('trace_lineno={}'.format(inspect.getsourcelines(fn)[1]))
    except IndexError:
        pass
    return trace_info_list


def _get_request_query_args():
    traced_request_args = []
    for k, v in request.args.items():
        v = '********' if k in {'password', 'secret'} else __r.repr(v).replace('"', "'")
        traced_request_args.append('{}={}'.format(k, v))
    return traced_request_args


def _get_request_post_values():
    traced_post_values = []
    if request.method in {'POST', 'PUT'}:
        json_args = request.json
        if json_args is not None:
            for k, v in json_args.items():
                v = '********' if k in {'password', 'secret'} else __r.repr(v).replace('"', "'")
                traced_post_values.append('{}={}'.format(k, v))
        else:
            for k, v in request.form.items():
                v = '********' if k in {'password', 'secret'} else __r.repr(v).replace('"', "'")
                traced_post_values.append('{}={}'.format(k, v))
    return traced_post_values


def get_request_trace_info():
    trace_uuid = getattr(g, config.log_request_id)
    query_args = ' '.join(_get_request_query_args())
    post_values = ' '.join(_get_request_post_values())
    return 'trace_uuid={} endpoint={} method={} ' \
        ' query_args:{} post_values:{}'.format(
            trace_uuid, request.path, request.method,
            query_args, post_values)


def get_response_trace_info(response):
    trace_uuid = getattr(g, config.log_request_id)
    return 'trace_uuid={} ip={} method={} ' \
           ' path:{} status_code:{}'.format(
            trace_uuid, request.remote_addr, request.method,
            request.path, response.status_code)
